#ifndef Yum_h
#define Yum_h

/*
** DANGER DANGER!  If you #include this, it must be #include'd first.  This is
** so that "Python.h" will be included first and so that its <Python.h> will be
** included first.  Python obnoxiously demands that that header be included
** first, and things sometimes go awry if it isn't.  (For example, you may get
** duplicate definitions of _POSIX_C_SOURCE.)
*/

#include "Python.h"
#include "Package.h"
#include "Log.h"
#include <QWidget>
#include <QHash>
#include <QList>
#include <QString>
#include <QSet>
#include <QEventLoop>
#include "Progress.h"
#include "YumIODialogs.h"
#include "YumIOCapturers.h"
#include "YumReposUsed.h"
#include <signal.h>

class CollectThread;
class ApplyThread;
class CollectFileListsThread;
class CollectOtherDataThread;
class ClearEntriesThread;

class YumIO : public QObject, private Python {
   Q_OBJECT
public:
   explicit YumIO();
   ~YumIO();
public:
   void  setup(QWidget*       parent,
               HProgressBar*  totalProgressBar,
               HProgressBar*  itemProgressBar);
public:
   bool  initialize             ();
   bool  collect                (const YumReposUsed&  reposUsed);
   bool  collectFileLists       ();
   bool  collectOtherData       ();
   void  reset                  ();
   void  clearEntries           ();  
   void  clearEntriesBeforeQuit ();
   bool  apply                  ();
public:
   bool  cleanDbcache  ();
   bool  cleanPackages ();
   bool  cleanMetadata ();
   bool  cleanAll      ();
protected:
   bool  cleanType (const char*  name,
                    const char*  functionName);
public:
   inline int  count() const 
   { if (clearing) return 0; else return list.count(); }
   inline Package::EntryList::const_iterator  begin() const
   { if (clearing) return emptyList.begin(); else return list.begin(); }
   inline Package::EntryList::const_iterator  end() const 
   { if (clearing) return emptyList.end(); else return list.end(); }
   inline Package::Entry* const&  operator[](int i) const
   { if (clearing) return emptyList[0]; else return list[i]; }
public:
   QString  getState            (const Package::Entry&  entry);
   QString  getKind             (const Package::Entry&  entry);
   // For packages & suites
   QString  getName             (const Package::Entry&  entry)
      { return entry.name; }
   QString  getVerRel           (const Package::Entry&  entry)
      { return entry.verrel; }
   QString  getRepo             (const Package::Entry&  entry)
      { return entry.repo; }
   QString  getArch             (const Package::Entry&  entry)
      { return entry.arch; }
   QString  getSize             (const Package::Entry&  entry)
      { return entry.sizeText; }
   QString  getSummary          (const Package::Entry&  entry)
      { return entry.summary; }
   QString  getDescription      (const Package::Entry&  entry);
   QString  getGroup            (const Package::Entry&  entry);
   QString  getVendor           (const Package::Entry&  entry);
   QString  getBuildTime        (const Package::Entry&  entry)
      { return entry.buildTimeText; }
   QString  getBuildHost        (const Package::Entry&  entry);
   QString  getInstallTime      (const Package::Entry&  entry);
   QString  getBinaryRpm        (const Package::Entry&  entry);
   QString  getSourceRpm        (const Package::Entry&  entry);
   QString  getLicense          (const Package::Entry&  entry);
   QString  getPackager         (const Package::Entry&  entry);
   QString  getUrl              (const Package::Entry&  entry);
   QString  getRelocations      (const Package::Entry&  entry);
   QString  getProvides         (const Package::Entry&  entry);
   QString  getRequires         (const Package::Entry&  entry);
   QString  getRequiresPre      (const Package::Entry&  entry);
   QString  getConflicts        (const Package::Entry&  entry);
   QString  getObsoletes        (const Package::Entry&  entry);
   QString  getRecommends       (const Package::Entry&  entry);
   QString  getSuggests         (const Package::Entry&  entry);
   QString  getSupplements      (const Package::Entry&  entry);
   QString  getEnhances         (const Package::Entry&  entry);
   QString  getFiles            (const Package::Entry&  entry);
   QString  getChangeLog        (const Package::Entry&  entry);
   // For groups
   QString  getMandatoryPkgs    (const Package::Entry&  entry);
   QString  getDefaultPkgs      (const Package::Entry&  entry);
   QString  getOptionalPkgs     (const Package::Entry&  entry);
   QString  getConditionalPkgs  (const Package::Entry&  entry);
public:
   QString  getYumVar (QString  name);
public:
   class EntryKey;
protected:
   typedef QHash<EntryKey,Package::Entry*>  EntryHash;
   Package::EntryList  list;
   Package::EntryList  emptyList;
   EntryHash           hash;
   QString             basearch;
   typedef QSet<QString>  StringSet;
   typedef QHash<QString,Package::Entry*>  EntryNameHash;
   EntryNameHash       namesBest;
   StringSet           biarches;
   bool                clearing;
protected:
   QStringList      collectRepos      (PyObject*              dict);
   void             useRepos          (PyObject*              repos,
                                       const YumReposUsed&    reposUsed);
   bool             collectField      (EntryHash&             hash,
                                       PyObject*              info,
                                       const char*            field,
                                       Package::State         state);
   bool             collectField      (EntryHash&             hash,
                                       PyObject*              info,
                                       char*                  field,
                                       Package::State         state);
   bool             collectSet        (EntryHash&             hash,
                                       PyObject*              set,
                                       Package::State         state);
   bool             collectPkg        (Package::Entry*&       entry,
                                       Package::State         state,
                                       PyObject*              pkg);
   bool             collectGroups     (EntryHash&             hash,
                                       PyObject*              groupInfo);
   bool             collectGroup      (Package::Entry*&       entry,
                                       PyObject*              groupObj,
                                       QString                name,
                                       PyObject*              historyGroup);
   void             clearEntriesReal  ();
   bool             upToDateGroup     (const Package::Entry&  entry);
   bool             upToDateGroupPkgs (const Package::Entry&  entry,
                                       const char*            attr);
   bool             upToDateGroupPkgs (const Package::Entry&  entry,
                                       char*                  attr);
   const char*      decodePrcoFlags   (PyObject*              flags);
   QString          getRelDepList     (PyObject*              pkg,
                                       const char*            key);
   QString          getRpmList        (PyObject*              pkg,
                                       const char*            rpmKey);
   QString          getGroupPkgs      (const Package::Entry&  entry,
                                       const char*            attr);
   QString          makeSorted        (QStringList&           list);
   QString          makeSortedUniq    (QStringList&           list);
   PyObject*        getRpmHdrValue    (PyObject*              pkg,
                                       const char*            hdrKey);
   bool             verifyLocalPkg    (PyObject*              pkg);
   Package::Kind    getPkgKind        (PyObject*              pkg);
   PyObject*        getGroupItem      (const Package::Entry&  entry,
                                       const char*            attr);
   Package::Entry*  findPackage       (QString                name);
   bool             applyGroups       ();
   typedef bool  (YumIO::*ApplyGrpChkFunc)(const Package::Entry*  grp,
                                           Package::Entry*        pkg);
   bool             applyGrpPkgs      (const Package::Entry*  entry,
                                       const char*            attr,
                                       ApplyGrpChkFunc        func);
   bool             applyGrpChkRemove (const Package::Entry*  grp,
                                       Package::Entry*        pkg);
   bool             applyGrpChkInsUpd (const Package::Entry*  grp,
                                       Package::Entry*        pkg);
   bool             applyGrpConflict  (const Package::Entry*  grp,
                                       Package::Entry*        pkg);
   QString          simpleString      (PyObject*              pkg,
                                       const char*            key);
   QString          getVerRel         (PyObject*              pkg);
   QString          getVerRelTxitem   (PyObject*              pkg);
   QString          getBiArch         (QString                arch);
   Package::Size    sizeOf            (PyObject*              pkg);
   QString          bytesFormat       (Package::Size          bytes);
   time_t           buildTimeOf       (PyObject*              pkg);
   char*            dateFormat        (time_t                 secs);
   char*            timeFormat        (time_t                 secs);
   QString          dnfVersionStr     ();
   long             splitVersion      (QString  str);
   long             dnfVersion        ();
   QString          pyString          (PyObject*              obj);
   QString          pyStringDec       (PyObject*              obj);
   QString          pyStringX         (PyObject*              obj);
   QString          pyStringXDec      (PyObject*              obj);
   long             pyLongDec         (PyObject*              obj);
   QString          pyMappingString   (PyObject*              obj,
                                       const char*            attr);
   time_t           pyTime            (PyObject*              obj);
   time_t           pyTimeDec         (PyObject*              obj);
   QString          pyDate            (PyObject*              obj);
   PyObject*        pyDictGetItem     (PyObject*              dict,
                                       const QString&         key);
   const char*      pyDescribe        (PyObject*              obj);
   void             log               (Log::Kind              kind,
                                       unsigned int           value,
                                       QString                text);
   bool             reapChildren      ();
   void             sigchldDisable    ();
   void             sigchldRestore    ();
   virtual void     pyExceptExpect    (PyObject*              xcpt);
   virtual void     pyExceptClear     ();
   void             pythonException   (QString                text);
   void             pythonExceptionV  (QString                format);
   void             commExceptSetup   ();
   bool             commExceptRaised  (PyObject*              result,
                                       QString                descr);
protected slots:
   void  blockerRelease ();
protected:
   void  execWorkThread       (QEventLoop&          eventLoop,
                               QThread&             thread);
   bool  collectReal          (const YumReposUsed&  reposUsed);
   bool  applyReal            ();
   bool  collectFileListsReal ();
   bool  collectOtherDataReal ();
   friend class CollectThread;
   friend class ApplyThread;
   friend class CollectFileListsThread;
   friend class CollectOtherDataThread;
   friend class ClearEntriesThread;
public:
   // These are public because they're callbacks from YumIOC.
   PyObject*  stdoutWrite                 (PyObject*  self, PyObject*  args);
   PyObject*  stderrWrite                 (PyObject*  self, PyObject*  args);
   PyObject*  log                         (PyObject*  self, PyObject*  args);
   PyObject*  errorlog                    (PyObject*  self, PyObject*  args);
   PyObject*  downloadStart               (PyObject*  self, PyObject*  args);
   PyObject*  downloadProgress            (PyObject*  self, PyObject*  args);
   PyObject*  downloadMessage             (PyObject*  self, PyObject*  args);
   PyObject*  downloadEnd                 (PyObject*  self, PyObject*  args);
   PyObject*  listTransaction             (PyObject*  self, PyObject*  args);
   PyObject*  reportDownloadSize          (PyObject*  self, PyObject*  args);
   PyObject*  reportRemoveSize            (PyObject*  self, PyObject*  args);
   PyObject*  userconfirm                 (PyObject*  self, PyObject*  args);
   PyObject*  downloadRemotePayloads      (PyObject*  self, PyObject*  args);
   PyObject*  rpmtransInit                (PyObject*  self, PyObject*  args);
   PyObject*  rpmtransDel                 (PyObject*  self, PyObject*  args);
   PyObject*  rpmtransProgress            (PyObject*  self, PyObject*  args);
   PyObject*  rpmtransScriptout           (PyObject*  self, PyObject*  args);
   PyObject*  postTransactionOutput       (PyObject*  self, PyObject*  args);
   PyObject*  gpgsigcheck                 (PyObject*  self, PyObject*  args);
protected:
   void  collectStart    (long long        totalFiles,
                          long long        totalSize,
                          long long        totalDrpms);
   void  collectProgress (PyObject*        payload,
                          QString          name,
                          Progress::Steps  size,
                          Progress::Steps  itemProgress);
   void  collectMessage  (QString          msg);
   void  collectEnd      (PyObject*        payload,
                          QString          name,
                          Progress::Steps  size,
                          long             status,
                          QString          msg);
public:
   void             stdoutCapturedWrite (QString  content);
   void             stderrCapturedWrite (QString  content);
protected:
   PyObject*        stdWrite           (PyObject*        self,
                                        PyObject*        args,
                                        Log::Kind        kind,
                                        unsigned int     value,
                                        QString&         buffer);
   void             write              (QString          content,
                                        Log::Kind        kind,
                                        unsigned int     value,
                                        QString&         buffer);
   void             exceptionExpected  ();
protected:
   QString          txitemFormat       (PyObject*        txitem);
   QString          txitemRepoFormat   (PyObject*        txitem);
   void             tsDownloadStart    (long long        totalFiles,
                                        long long        totalSize,
                                        long long        totalDrpms);
   void             tsDownloadProgress (PyObject*        payload,
                                        QString          name,
                                        Progress::Steps  size,
                                        Progress::Steps  itemProgress);
   void             tsDownloadMessage  (QString          msg);
   void             tsDownloadEnd      (PyObject*        payload,
                                        QString          name,
                                        Progress::Steps  size,
                                        long             status,
                                        QString          msg);
   PyObject*        findRelation       (PyObject*        relatedto,
                                        const char*      relation);
   Package::Entry*  installedEntry     (Package::Entry*  entry);
signals:
   void             collectedExtraData();
protected:
   void error   (const char*  text);
   void warning (const char*  text);
public:
   void error   (QString  text);
   void warning (QString  text);
public:
   Log::EntryList     logEntries;
   int                logStartOp;
   bool               collectedFileLists;
   bool               collectedOtherData;
protected:
   FILE*              file_dbg;
   QWidget*           parent;
   YumIODialogs*      dialogs;
   YumIOCapturers*    capturers;
   PyObject*          rpm_dict;
   PyObject*          classYum;
   PyObject*          yumIO;
   PyObject*          base;
   PyObject*          reposDict;
   bool               locked;
   bool               yumVerKnown;
   QString            yumVer;
   bool               dnfVerKnown;
   QString            dnfVer;
   bool               usesPrestoKnown;
   bool               usesPrestoValue;
   struct {
      PyObject*       bdbQuit;
      PyObject*       yumBaseError;
      PyObject*       miscError;
      PyObject*       repoError;
      PyObject*       sqlite3OperationalError;
      PyObject*       lockError;
      PyObject*       noMoreMirrorsRepoError;
   } xcpt;
   struct {
      PyObject*       deltaPayload;
   } types;
   struct {
      QString         stdoutBuffer;
      QString         stderrBuffer;
      Log::Kind       stderrKind;
      unsigned int    stderrValue;
   } wr;
   HProgressBar*      totalProgressBar;
   HProgressBar*      itemProgressBar;
   Progress::Object*  pg;
   // Mode for downloads
   typedef enum {
      Undefined,
      Collect,
      Ts,
   } Mode;
   struct {
      Mode  mode;
   } download;
   // Progress for reading repositories (collect)
   struct {
      QStringList  repos;
      QString      kind;
      int          act;
      int          total;
      QString      name;
   } repoProgress;
   // State for transaction (apply)
   typedef QHash<QString,Progress::Threads>  TsMeterMapping;
   struct {
      // descr holds the description for the current user prompt.  It is
      // populated in one of two ways:
      //
      // list.during:
      //    This is the normal case, and it is populated only with NUU-specified
      //    text.
      // !list.during:
      //    The rest of the time, it accumulates arbitrary logged output, just
      //    in case yum is planning to throw out some unexpected confirmation
      //    query.  After each userconfirm, the text is cleared so that any
      //    subsequent userconfirm is uncluttered.  Also, there are a few places
      //    where we know that subsequent logged text is about to start a known
      //    query, and the text is cleared then so that it is uncluttered.
      QString  descr;
      // State for listing the transaction to the user and confirming it
      struct {
         bool  during;
         bool  confirmed;
      } list;
      unsigned int   act;  
      // State for running transaction (including rpm grabs)
      struct {
         bool           doingTest;           // doing test (as opposed to run)
         Package::Size  localSize;           // size of local cached rpms
         Package::Size  downloadSize;        // size of rpms to be downloaded
         Package::Size  nominalDownloadProg; // progress needed for rpm downloads
         Package::Size  actualDownloadProg;  // actual progress needed after
                                             // cached rpms & drpms logic
         unsigned int   actualDownloadCount; // actual number of rpm downloads
         unsigned int   pgRDownloadRpms;     // total progress for rpm downloads
         Package::Size  totalTsSize;         // size of installs & erases
         unsigned long  lastTs;              // last rpmTransProgress currentTs
         long           lastAction;          // last rpmTransProgress action
         bool           deltaRebuilt;        // <delta rebuild> happened
      } run;
      // State for parallel grabs (either hdr or rpm grabs, depending on
      // ts.confirmed).
      struct {
         bool               begun;
         long long          files;
         Package::Size      size;
         Progress::Threads  threadsStarted;
         TsMeterMapping     meterMapping;
      } parallel;
   } ts;
   struct sigaction  savedSigChldAct;
   class Blocker : QMutex {
   public:
      Blocker() : QMutex() {}
      void prepare() { lock(); }           /* called by creator thread */
      void release() { unlock(); }         /* called by creator thread */
      void block()   { lock(); unlock(); } /* called by created thread */
   };
   Blocker blocker;
};

#endif
